/****************************************
Filename: Util/DRTime.h
Description: Videotester Time operations
Author: Christian Linhart
Copyright: (C)1997 Christian Linhart
****************************************/

#ifndef VtTime_h
#define VtTime_h

#include <sys/time.h>

#include <X11/Xfuncproto.h>
_XFUNCPROTOBEGIN


/* Data structures */

/* Time difference, may be negative */
struct timediff{
    int td_sec;
    int td_usec;
};

/* Time difference split into common human-used  units hours:minutes:seconds:milliseconds */
struct timediff_expanded{
    int fHours;
    int fMinutes;
    int fSeconds;
    int fMilliseconds;
    int fMicroseconds; /* within one millisecond */
    int fTotalMicroseconds; /* within one second */
    struct timediff fTimeDiff; /* the original time diff */
};


/* Create timetiff with specified value */
struct timediff VtTimeDiffNull(void);

struct timediff VtTimeDiffUsecs(int usecs);
struct timediff VtTimeDiffUsecsDouble(double usecs);

struct timediff VtTimeDiffMsecs(int msecs);
struct timediff VtTimeDiffMsecsDouble(double msecs);

struct timediff VtTimeDiffSecs(int secs);
struct timediff VtTimeDiffSecsDouble(double secs);

struct timediff VtTimeDiffMinutes(int minutes);
struct timediff VtTimeDiffMinutesDouble(double minutes);

struct timediff VtTimeDiffHours(int hours);
struct timediff VtTimeDiffHoursDouble(double hours);

struct timediff VtTimeDiffDays(int days);
struct timediff VtTimeDiffDaysDouble(double days);

struct timediff VtTimeDiffWeeks(int weeks);
struct timediff VtTimeDiffWeeksDouble(double weeks);

/* Conversion between Time and Timediff */
struct timediff VtTimeToTimeDiff(struct timeval t);
struct timeval VtTimeDiffToTime(struct timediff d);

/* Convert timediff to various units */
/* microseconds */
int VtTimeDiffToUsecs(struct timediff d, Bool *overflow);
double VtTimeDiffToUsecsDouble(struct timediff d);
/* milliseconds */
int VtTimeDiffToMsecs(struct timediff d, Bool *overflow);
double VtTimeDiffToMsecsDouble(struct timediff d);
/* seconds */
int VtTimeDiffToSecs(struct timediff d, Bool *overflow);
double VtTimeDiffToSecsDouble(struct timediff d);
/* Minutes */
int VtTimeDiffToMinutes(struct timediff d, Bool *overflow);
double VtTimeDiffToMinutesDouble(struct timediff d);
/* Hours */
int VtTimeDiffToHours(struct timediff d, Bool *overflow);
double VtTimeDiffToHoursDouble(struct timediff d);
/* Days */
int VtTimeDiffToDays(struct timediff d, Bool *overflow);
double VtTimeDiffToDaysDouble(struct timediff d);
/* Weeks */
int VtTimeDiffToWeeks(struct timediff d, Bool *overflow);
double VtTimeDiffToWeeksDouble(struct timediff d);

/* timespec */
struct timespec VtTimeDiffToTimespec(struct timediff d);

/* expanded */
void VtTimeDiffToExpanded(
    struct timediff d, struct timediff_expanded *expanded);


/* Arithmetic operations */
struct timeval VtTimeAdd(struct timeval t, struct timediff d);
struct timediff VtTimeAddDiffs(
    struct timediff d1, struct timediff d2);
struct timeval VtTimeSub(struct timeval t, struct timediff d);
struct timediff VtTimeSubDiffs(
    struct timediff d1, struct timediff d2);
struct timediff VtTimeDiff(struct timeval t1, struct timeval t2);
struct timediff VtTimeMul(double factor, struct timediff d);
double VtTimeDiv_Double( struct timediff d1, struct timediff d2 );
struct timediff VtTimeMod( struct timediff d1, struct timediff d2 );

/* Change sign */
struct timediff VtTimeDiffMinus(struct timediff d);

/* Comparison */
#define VtTimeLess -1
#define VtTimeEqual 0
#define VtTimeGreater 1
int VtTimeCompare(struct timeval t1, struct timeval t2);
int VtTimeCompareDiff(struct timediff d1, struct timediff d2);

/* Get current time (return Null-time in case of error) */
enum ETimePrecision{
	eTimePrecision_Exact,
	eTimePrecision_Tick,
	eTimePrecision_10msecFast,
	eTimePrecision_10msecFastThreadSafe /* only available in CppFramework */
};

struct timeval VtTimeNow(void);

struct timeval VtTimeNowWithPrecision(enum ETimePrecision precision);

/* Get null time  (UNIX epoch) */
struct timeval VtTimeNull(void);

Bool VtTimeIsNull( struct timeval t );

struct timediff VtTimeDiffZero(void);

/* Wait for specified period of time */
void VtTimeDelay(struct timediff d);

_XFUNCPROTOEND

#endif
