/****************************************
Filename: lib/DispRecXtAPI/DispRecXtAPI.c
Description: API for DisplayRecorder/DemoRecorder integrated into Xt
Author: Christian Linhart
Copyright: (C)2006 Christian Linhart
****************************************/

#include "DispRecXtAPI.h"


#include <X11/Intrinsic.h>
#if 0
#include "DispRec/error.h"
#include "DispRec/DispRecCtrlExtension.h"
#else
#include "DispRec/DRError.h"
#include "DispRec/DRMem.h"
#include "DispRec/DispRecCtrlExtension.h"
#endif

/*****************************************
******************************************
* Internal Types
******************************************
*****************************************/

/**
* Object which identifies a connection to a DisplayRecorder/DemoRecorder X11-Proxy.
*/
struct DispRecXtAPI_Server_struct{
    Display* fDisplay;
    int fDispRecCtrl_event_base;
    int fDispRecCtrl_error_base;
    
    int fDispatchProcId;
    
    DispRecXtAPI_RecordingSession fSessions;
    DispRecXtAPI_GUI fGUIs;
    
    DispRecXtAPI_Server fNext;
    DispRecXtAPI_Server* fPrev;
};

/**
* A recording session is an object which can be used to condigure, start and stop a recording.
*/
struct DispRecXtAPI_RecordingSession_struct{
    DispRecCtrl_RecordingSession fExtSession;
    
    /* config */
    unsigned int fLength_recording_time;  /* in seconds, 0 means unlimited */
    Bool fEndless;
	/* True: endless recording, only the last length seconds are
		guaranteed to be recorded. (some more time may be recorded due to
		the way endless recording is implemented)
	    False: stop recording after max_length seconds
	*/
    Bool fDoSoundRecording;
    Bool fMuteSoundRecording;
    Bool fFinishOtherRecordingsBeforeStart;
    
    /* callbacks */
    DispRecXtAPI_RecordingStarted_CallbackHandle fRecordingStarted_Callbacks;
    DispRecXtAPI_RecordingNotStarted_CallbackHandle fRecordingNotStarted_Callbacks;
    DispRecXtAPI_RecordingFinished_CallbackHandle fRecordingFinished_Callbacks;
    DispRecXtAPI_RecordingNotFinished_CallbackHandle fRecordingNotFinished_Callbacks;
    DispRecXtAPI_RecLevel_CallbackHandle fRecLevel_Callbacks;
    DispRecXtAPI_RecordingSessionDeleted_CallbackHandle fRecordingSessionDeleted_Callbacks;
    
    /* ownership */
    Bool fWeAreTheOwner; /* if True, this program is the owner */
    
    /* structure: pointer to container and siblings in list */
    DispRecXtAPI_Server fServer;
    DispRecXtAPI_RecordingSession fNext;
    DispRecXtAPI_RecordingSession *fPrev;
};

/**
* DispRecXtAPI_GUI is a handle for communicating with a DemoRecorder GUI
*/
struct DispRecXtAPI_GUI_struct{
    DispRecCtrl_GUI fExtGui;

    /* callbacks */
    DispRecXtAPI_GUIStarted_CallbackHandle fGUIStarted_Callbacks;
    DispRecXtAPI_GUINotStarted_CallbackHandle fGUINotStarted_Callbacks;
    DispRecXtAPI_GUIExited_CallbackHandle fGUIExited_Callbacks;
    DispRecXtAPI_GUIRestarting_CallbackHandle fGUIRestarting_Callbacks;
    DispRecXtAPI_GUIRestarted_CallbackHandle fGUIRestarted_Callbacks;
    DispRecXtAPI_GUINotRestarted_CallbackHandle fGUINotRestarted_Callbacks;
    DispRecXtAPI_GUIVisible_CallbackHandle fGUIVisible_Callbacks;
    DispRecXtAPI_GUIHidden_CallbackHandle fGUIHidden_Callbacks;
    DispRecXtAPI_GUIControlsRecordingSession_CallbackHandle fGUIControlsRecordingSession_Callbacks;
    DispRecXtAPI_GUILeavesRecordingSession_CallbackHandle fGUILeavesRecordingSession_Callbacks;
    DispRecXtAPI_GUIHandleDeleted_CallbackHandle fGUIHandleDeleted_Callbacks;

    /* structure: pointer to container and siblings in list */
    DispRecXtAPI_Server fServer;
    DispRecXtAPI_GUI fNext;
    DispRecXtAPI_GUI *fPrev;
};

/*****************************************
******************************************
* Internal Data
******************************************
*****************************************/

static XtAppContext DispRecXtAPI_app_context;

static DispRecXtAPI_Server DispRecXtAPI_servers = NULL;
    
/*****************************************
******************************************
callback handles with register and deregister functions
******************************************
*****************************************/

#define DispRecXtAPI_CallbackHandleDef( containerType, name ) \
    struct DispRecXtAPI_##name##_CallbackHandle_struct{ \
	DispRecXtAPI_##name##_CallbackProc fProc; \
	void* fClientData; \
	containerType fContainer; \
	\
	DispRecXtAPI_##name##_CallbackHandle fNext; \
	DispRecXtAPI_##name##_CallbackHandle *fPrev; \
    }; \
    \
    DispRecXtAPI_##name##_CallbackHandle gMatchAll_##name##_Callbacks = NULL; \
    \
    DispRecXtAPI_##name##_CallbackHandle \
	DispRecXtAPI_Add_##name##_Callback( \
	    containerType container, \
	    DispRecXtAPI_##name##_CallbackProc proc, \
	    void *client_data, \
	    VtError *err \
	) \
	{ \
	    DispRecXtAPI_##name##_CallbackHandle handle = \
		NEW( struct DispRecXtAPI_##name##_CallbackHandle_struct ); \
	    handle->fProc = proc; \
	    handle->fClientData = client_data; \
	    handle->fContainer = container; \
	    if ( container == containerType##_None ) { \
		/* for all containers */ \
		handle->fNext = gMatchAll_##name##_Callbacks; \
		if ( handle->fNext ) { \
		    handle->fNext->fPrev = &( handle->fNext ); \
		} \
		gMatchAll_##name##_Callbacks = handle; \
		handle->fPrev = &( gMatchAll_##name##_Callbacks ); \
		 \
	    } else { \
		/* for the given container */ \
		 \
		handle->fNext = handle->fContainer->f##name##_Callbacks; \
		if ( handle->fNext ) { \
		    handle->fNext->fPrev = &( handle->fNext ); \
		} \
		handle->fContainer->f##name##_Callbacks = handle; \
		handle->fPrev = &( handle->fContainer->f##name##_Callbacks ); \
	    } \
	    \
	    return handle; \
	} \
	\
    void DispRecXtAPI_Remove_##name##_Callback( \
	DispRecXtAPI_##name##_CallbackHandle handle, \
	VtError *err \
    ) { \
	*(handle->fPrev) = handle->fNext; \
	if ( handle->fNext ) { \
	    handle->fNext->fPrev = handle->fPrev; \
	} \
	handle->fNext = NULL; \
	handle->fPrev = NULL; \
	DELETE( handle ); \
    } \
        
/* dispatch a callback. 
    argsList has to be the full argslist including parentheses.. 
    the callback-handle is the parameter "handle". 
    So it has to have the last arg "handle->fClientData", etc  */
#define DispRecXtAPI_CallCallbacks( containerType, name, container, argsList ) \
    \
    { \
	DispRecXtAPI_##name##_CallbackHandle handle; \
	for( \
	    handle =  gMatchAll_##name##_Callbacks; \
	    handle != NULL; \
	    handle = handle->fNext \
	) { \
	    handle->fProc argsList; \
	} \
	if ( container != containerType##_None ) { \
	    for( \
		handle =  container->f##name##_Callbacks; \
		handle != NULL; \
		handle = handle->fNext \
	    ) { \
		handle->fProc argsList; \
	    } \
	} \
    }


DispRecXtAPI_CallbackHandleDef( 
    DispRecXtAPI_RecordingSession, 
    RecordingStarted 
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_RecordingSession,
    RecordingNotStarted
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_RecordingSession,
    RecordingFinished
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_RecordingSession,
    RecordingNotFinished
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_RecordingSession,
    RecLevel
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_RecordingSession,
    RecordingSessionDeleted
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,
    GUIStarted
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,	
    GUINotStarted
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,
    GUIRestarting
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,
    GUIRestarted
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,
    GUINotRestarted
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,	
    GUIExited
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,	
    GUIVisible
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,	
    GUIHidden
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,	
    GUIControlsRecordingSession
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,	
    GUILeavesRecordingSession
)

DispRecXtAPI_CallbackHandleDef(
    DispRecXtAPI_GUI,	
    GUIHandleDeleted
)
    
/*****************************************
******************************************
* Functions   
******************************************
*****************************************/

/*****************************************
enum to string conversion functions
*****************************************/

#define CASE( prefix, value ) \
    case prefix##value: return #value;


const char* DispRecXtAPI_RecordingNotStartedReason_ToString( 
    enum DispRecXtAPI_RecordingNotStartedReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_RecordingNotStartedReason_, AlreadyRecording )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, CannotOpenFile )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, MaxLengthReached )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, MaxSizeReached )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, MaxFsUsageReached )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, BadValue )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, WriteError )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, FileDisappeared )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, TotalMaxLengthReached )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, NoLicense )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, SoundDeviceReadError )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, SoundDeviceConfigError )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, SoundDeviceOpenError )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, SoundDeviceMissingCapabilities )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, ChildprocessDiedUnexpectedly )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, CannotStartSoundCompression )
	CASE( DispRecXtAPI_RecordingNotStartedReason_, NoSoundSupport )
	default: 
	    return "illegal_value";
    }
}

const char* DispRecXtAPI_RecordingFinishedReason_ToString( 
    enum DispRecXtAPI_RecordingFinishedReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_RecordingFinishedReason_, ByClientRequest )
	CASE( DispRecXtAPI_RecordingFinishedReason_, CannotOpenFile )
	CASE( DispRecXtAPI_RecordingFinishedReason_, MaxLengthReached )
	CASE( DispRecXtAPI_RecordingFinishedReason_, MaxSizeReached )
	CASE( DispRecXtAPI_RecordingFinishedReason_, MaxFsUsageReached )
	CASE( DispRecXtAPI_RecordingFinishedReason_, BadValue )
	CASE( DispRecXtAPI_RecordingFinishedReason_, WriteError )
	CASE( DispRecXtAPI_RecordingFinishedReason_, FileDisappeared )
	CASE( DispRecXtAPI_RecordingFinishedReason_, TotalMaxLengthReached )
	CASE( DispRecXtAPI_RecordingFinishedReason_, NoLicense )
	CASE( DispRecXtAPI_RecordingFinishedReason_, SoundDeviceReadError )
	CASE( DispRecXtAPI_RecordingFinishedReason_, SoundDeviceConfigError )
	CASE( DispRecXtAPI_RecordingFinishedReason_, SoundDeviceOpenError )
	CASE( DispRecXtAPI_RecordingFinishedReason_, SoundDeviceMissingCapabilities )
	CASE( DispRecXtAPI_RecordingFinishedReason_, ChildprocessDiedUnexpectedly )
	CASE( DispRecXtAPI_RecordingFinishedReason_, CannotStartSoundCompression )
	CASE( DispRecXtAPI_RecordingFinishedReason_, NoSoundSupport )
	CASE( DispRecXtAPI_RecordingFinishedReason_, NotRecording )
	default: 
	    return "illegal_value";
    }
}

const char* DispRecXtAPI_RecordingNotFinishedReason_ToString( 
    enum DispRecXtAPI_RecordingNotFinishedReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_RecordingNotFinishedReason_, NotRecording )
	default: 
	    return "illegal_value";
    }
}

const char* DispRecXtAPI_RecordingSessionDeletedReason_ToString( 
    enum DispRecXtAPI_RecordingSessionDeletedReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_RecordingSessionDeletedReason_, ByClientRequest )
	default: 
	    return "illegal_value";
    }
}

const char* DispRecXtAPI_GUINotStartedReason_ToString(
    enum DispRecXtAPI_GUINotStartedReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_GUINotStartedReason_, CannotOpenFile )
	CASE( DispRecXtAPI_GUINotStartedReason_, BadValue )
	CASE( DispRecXtAPI_GUINotStartedReason_, NoLicense )
	CASE( DispRecXtAPI_GUINotStartedReason_, AlreadyRunning )
	default: 
	    return "illegal_value";
    }    
}

const char* DispRecXtAPI_GUIExitedReason_ToString(
    enum DispRecXtAPI_GUIExitedReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_GUIExitedReason_, ByClientRequest )
	CASE( DispRecXtAPI_GUIExitedReason_, NoLicense )
	CASE( DispRecXtAPI_GUIExitedReason_, Killed )
	default: 
	    return "illegal_value";
    }        
}

const char* DispRecXtAPI_GUIRestartingReason_ToString(
    enum DispRecXtAPI_GUIRestartingReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_GUIRestartingReason_, ProcessTerminated )
	default: 
	    return "illegal_value";
    }    
    
}

const char* DispRecXtAPI_GUINotRestartedReason_ToString(
    enum DispRecXtAPI_GUINotRestartedReason value
) {
    switch ( value ) {
	CASE( DispRecXtAPI_GUINotRestartedReason_, CannotOpenFile )
	CASE( DispRecXtAPI_GUINotRestartedReason_, BadValue )
	CASE( DispRecXtAPI_GUINotRestartedReason_, NoLicense )
	CASE( DispRecXtAPI_GUINotRestartedReason_, AlreadyRunning )
	default: 
	    return "illegal_value";
    }        
}

#undef CASE

/*****************************************
forward decl of Internal functions
*****************************************/

static Boolean DispRecXtAPI_Dispatch_DispRecCtrlExtEvent(
    DispRecXtAPI_Server server,
    XEvent* event);

/*****************************************
Internal functions
*****************************************/


static void DispRecXtAPI_RecordingSession_InitDataAndInsertIntoStructure(
	DispRecXtAPI_Server server,
	DispRecXtAPI_RecordingSession session,
	VtError *err
    )
{
    /* init lists of callbacks */
    session->fRecordingStarted_Callbacks = NULL;
    session->fRecordingNotStarted_Callbacks = NULL;
    session->fRecordingFinished_Callbacks = NULL;
    session->fRecordingNotFinished_Callbacks = NULL;
    session->fRecLevel_Callbacks = NULL;
    session->fRecordingSessionDeleted_Callbacks = NULL;
    
    /* insert into list of sessions */
    session->fNext = server->fSessions;
    if ( session->fNext ) {
	session->fNext->fPrev = &(session->fNext);
    }
    session->fPrev = &(server->fSessions);
    server->fSessions = session;
    
}

static DispRecXtAPI_RecordingSession
    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
	DispRecXtAPI_Server server,
	DispRecCtrl_RecordingSession ext_session,
	VtError *err
    )
{
    if ( ext_session == DispRecCtrl_RecordingSession_None ) {
	return DispRecXtAPI_RecordingSession_None;
    }
    {
	    /* search */
	    DispRecXtAPI_RecordingSession session = server->fSessions;
	    while( session != NULL ) {
		if ( session->fExtSession == ext_session ) {
		    return session;
		}
		session = session->fNext;
	    }
	    
	    /* not found --> create new session object ( with foreign ownership ) */
	    session = NEW( struct DispRecXtAPI_RecordingSession_struct );
	    session->fExtSession = ext_session;
	    /* ownership --> True because we created it */
	    session->fWeAreTheOwner = True;
	    /* TODO: change lower level api such that these properties are
		configured separately and not passed with starting of the recording
		anymore */
	    session->fLength_recording_time = 0; /* 0 means: unlimited */
	    session->fEndless = False;
	    session->fDoSoundRecording = True;
	    session->fMuteSoundRecording = False;
	    session->fFinishOtherRecordingsBeforeStart = False;
	    
	     DispRecXtAPI_RecordingSession_InitDataAndInsertIntoStructure(
		 server, session, err );
	    
	    return session;
    }
}


static void DispRecXtAPI_Gui_InitDataAndInsertIntoStructure(
	DispRecXtAPI_Server server,
	DispRecXtAPI_GUI gui,
	VtError *err
    )
{
    gui->fServer = server;
    
    gui->fGUIStarted_Callbacks = NULL;
    gui->fGUINotStarted_Callbacks = NULL;
    gui->fGUIExited_Callbacks = NULL;
    gui->fGUIRestarting_Callbacks = NULL;
    gui->fGUIRestarted_Callbacks = NULL;
    gui->fGUINotRestarted_Callbacks = NULL;
    gui->fGUIVisible_Callbacks = NULL;
    gui->fGUIHidden_Callbacks = NULL;
    gui->fGUIControlsRecordingSession_Callbacks = NULL;
    gui->fGUILeavesRecordingSession_Callbacks = NULL;
    gui->fGUIHandleDeleted_Callbacks = NULL;
    
    /* insert into list */
    gui->fNext = server->fGUIs;
    if ( gui->fNext ) {
	gui->fNext->fPrev = &(gui->fNext);
    }
    server->fGUIs = gui;
    gui->fPrev = &( server->fGUIs );    
}

static DispRecXtAPI_GUI
    DispRecXtAPI_ConvertFrom_ExtGui(
	DispRecXtAPI_Server server,
	DispRecCtrl_GUI ext_gui,
	VtError *err
    )
{
    if ( ext_gui == DispRecCtrl_GUI_None ) {
	return DispRecCtrl_GUI_None;
    }
    {
	    /* search */
	    DispRecXtAPI_GUI gui = server->fGUIs;
	    while( gui != NULL ) {
		if ( gui->fExtGui == ext_gui ) {
		    return gui;
		}
		gui = gui->fNext;
	    }
	    /* not found --> create new gui object ( with foreign ownership ) */
	    gui = NEW( struct DispRecXtAPI_GUI_struct );
	    gui->fExtGui = ext_gui;
	    
	     DispRecXtAPI_Gui_InitDataAndInsertIntoStructure(
		 server, gui, err );

	    return gui;
    }
}


static void DispRecXtAPI_Internal_DeleteRecordingSessionObject( 
    DispRecXtAPI_RecordingSession session, VtError *err )
{
    /* remove from list */
    *(session->fPrev) = session->fNext;
    if ( session->fNext ) {
	session->fNext->fPrev = session->fPrev;
    }
    session->fNext = NULL;
    session->fPrev = NULL;
    
    /* delete data object */
    DELETE( session );
}

#define MAX_SERVERS 16

/* there is no client_data parameter for XtSetEventDispatcher, so we have to workaround that
    with a fixed number of predefined functions. */

DispRecXtAPI_Server DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Server[MAX_SERVERS];

#define Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer( servNum ) \
static Boolean DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_##servNum( \
    XEvent* event \
) \
{ \
    return DispRecXtAPI_Dispatch_DispRecCtrlExtEvent( \
	DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Server[servNum], \
	event \
    ); \
}

Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(0)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(1)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(2)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(3)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(4)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(5)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(6)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(7)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(8)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(9)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(10)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(11)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(12)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(13)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(14)
Declare_DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_ForServer(15)

XtEventDispatchProc DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Proc[MAX_SERVERS]=
{
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_0,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_1,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_2,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_3,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_4,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_5,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_6,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_7,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_8,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_9,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_10,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_11,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_12,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_13,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_14,
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_15
};

/* event-dispatcher for the Xt Event loop */
static Boolean DispRecXtAPI_Dispatch_DispRecCtrlExtEvent(
    DispRecXtAPI_Server server,
    XEvent* event)
{
    DECLERR;
    int eventBase = server->fDispRecCtrl_event_base;
    if (
	event->type >= eventBase &&
	event->type < eventBase+DispRecCtrl_NumberEvents)
    {
	/* This event is from rpserv --> dispatch it */
	switch(event->type - eventBase) {
	    case DispRecCtrl_Notify: {
		XDispRecCtrl_NotifyEvent *notifyevent=
		    (XDispRecCtrl_NotifyEvent*)event;
		switch(notifyevent->subtype) {
		    case DispRecCtrl_NotifyStarted: {
			XDispRecCtrl_NotifyStartedEvent *notifystarted=
			    (XDispRecCtrl_NotifyStartedEvent*)notifyevent;
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, notifystarted->recordingSession, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_RecordingSession, 
			    RecordingStarted,
			    session,
			    ( session, handle, handle->fClientData )
			);	
			return True;
		    } break;
		    case DispRecCtrl_NotifyNotStarted: {
			XDispRecCtrl_NotifyNotStartedEvent *notifynotstarted=
			    (XDispRecCtrl_NotifyNotStartedEvent*)notifyevent;
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, notifynotstarted->recordingSession, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_RecordingSession, 
			    RecordingNotStarted,
			    session,
			    ( session, handle, notifynotstarted->reason, handle->fClientData )
			);	
			return True;
		    } break;
		    case DispRecCtrl_NotifyFinished: {
			XDispRecCtrl_NotifyFinishedEvent 
			    *finished=
			    (XDispRecCtrl_NotifyFinishedEvent*)
			    notifyevent;
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, finished->recordingSession, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_RecordingSession, 
			    RecordingFinished,
			    session,
			    ( session, handle, finished->reason, handle->fClientData )
			);	
			return True;
		    } break;
		    case DispRecCtrl_NotifyNotFinished: {
			XDispRecCtrl_NotifyNotFinishedEvent 
			    *notFinished=
			    (XDispRecCtrl_NotifyNotFinishedEvent*)
			    notifyevent;
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, notFinished->recordingSession, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_RecordingSession, 
			    RecordingNotFinished,
			    session,
			    ( session, handle, notFinished->reason, handle->fClientData )
			);	
			return True;
		    } break;
		    case DispRecCtrl_NotifyRecordingSessionDeleted: {
			XDispRecCtrl_NotifyRecordingSessionDeletedEvent 
			    *sessionDeleted=
			    (XDispRecCtrl_NotifyRecordingSessionDeletedEvent*)
			    notifyevent;
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, sessionDeleted->recordingSession, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_RecordingSession, 
			    RecordingSessionDeleted,
			    session,
			    ( session, handle, sessionDeleted->reason, handle->fClientData )
			);	
			/* delete session object */
			DispRecXtAPI_Internal_DeleteRecordingSessionObject( session, err );
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIStarted: {
			XDispRecCtrl_NotifyGUIStartedEvent
			    *guiStarted=
			    (XDispRecCtrl_NotifyGUIStartedEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiStarted->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIStarted,
			    gui,
			    ( gui, handle, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUINotStarted: {
			XDispRecCtrl_NotifyGUINotStartedEvent
			    *guiNotStarted=
			    (XDispRecCtrl_NotifyGUINotStartedEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiNotStarted->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUINotStarted,
			    gui,
			    ( gui, handle, guiNotStarted->reason, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIExited: {
			XDispRecCtrl_NotifyGUIExitedEvent
			    *guiExited=
			    (XDispRecCtrl_NotifyGUIExitedEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiExited->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIExited,
			    gui,
			    ( gui, handle, guiExited->reason, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIRestarting: {
			XDispRecCtrl_NotifyGUIRestartingEvent
			    *guiRestarting=
			    (XDispRecCtrl_NotifyGUIRestartingEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiRestarting->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIRestarting,
			    gui,
			    ( gui, handle, guiRestarting->reason, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIRestarted: {
			XDispRecCtrl_NotifyGUIRestartedEvent
			    *guiRestarted=
			    (XDispRecCtrl_NotifyGUIRestartedEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiRestarted->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIRestarted,
			    gui,
			    ( gui, handle, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUINotRestarted: {
			XDispRecCtrl_NotifyGUINotRestartedEvent
			    *guiNotRestarted=
			    (XDispRecCtrl_NotifyGUINotRestartedEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiNotRestarted->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUINotRestarted,
			    gui,
			    ( gui, handle, guiNotRestarted->reason, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIVisible: {
			XDispRecCtrl_NotifyGUIVisibleEvent
			    *guiVisible=
			    (XDispRecCtrl_NotifyGUIVisibleEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiVisible->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIVisible,
			    gui,
			    ( 
				gui, handle, 
				guiVisible->screen, guiVisible->x, guiVisible->y,
				handle->fClientData 
			    )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIHidden: {
			XDispRecCtrl_NotifyGUIHiddenEvent
			    *guiHidden=
			    (XDispRecCtrl_NotifyGUIHiddenEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiHidden->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIHidden,
			    gui,
			    ( gui, handle, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIControlsRecordingSession: {
			XDispRecCtrl_NotifyGUIControlsRecordingSessionEvent
			    *guiControlsRecordingSession=
			    (XDispRecCtrl_NotifyGUIControlsRecordingSessionEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiControlsRecordingSession->gui, err 
			    );
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, guiControlsRecordingSession->recordingSession, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIControlsRecordingSession,
			    gui,
			    ( gui, handle, session, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUILeavesRecordingSession: {
			XDispRecCtrl_NotifyGUILeavesRecordingSessionEvent
			    *guiLeavesRecordingSession=
			    (XDispRecCtrl_NotifyGUILeavesRecordingSessionEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiLeavesRecordingSession->gui, err 
			    );
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, guiLeavesRecordingSession->recordingSession, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUILeavesRecordingSession,
			    gui,
			    ( gui, handle, session, handle->fClientData )
			);				    
			return True;
		    } break;
		    case DispRecCtrl_NotifyGUIHandleDeleted: {
			XDispRecCtrl_NotifyGUIHandleDeletedEvent
			    *guiHandleDeleted=
			    (XDispRecCtrl_NotifyGUIHandleDeletedEvent*)
			    notifyevent;
			DispRecXtAPI_GUI gui = 
			    DispRecXtAPI_ConvertFrom_ExtGui(
				server, guiHandleDeleted->gui, err 
			    );
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_GUI, 
			    GUIHandleDeleted,
			    gui,
			    ( gui, handle, handle->fClientData )
			);				    
			return True;
		    } break;
		} /* end switch subtype */
	    } break;
	    case DispRecCtrl_UserColor: {
		XDispRecCtrl_UserColorEvent *user_color_event=
		    (XDispRecCtrl_UserColorEvent*)event;
		switch(user_color_event->subtype) {
		    case DispRecCtrl_UserColor_ChangePixel: {
			XDispRecCtrl_UserColor_ChangePixelEvent 
			    *change_pixel=
			    (XDispRecCtrl_UserColor_ChangePixelEvent*)
			    user_color_event;
			/* not callbacks yet --> do nothing and return False */
			return False;
		    } break;
		    case DispRecCtrl_UserColor_Delete: {
			XDispRecCtrl_UserColor_DeleteEvent 
			    *delete_usercolor=
			    (XDispRecCtrl_UserColor_DeleteEvent*)
			    user_color_event;
			/* not callbacks yet --> do nothing and return False */
			return False;
		    } break;
		} /* end switch subtype */
	    } break;
	    case DispRecCtrl_DRSoundRec: {
		XDispRecCtrl_DRSoundRecEvent *drSoundRec_event=
		    (XDispRecCtrl_DRSoundRecEvent*)event;
		switch(drSoundRec_event->subtype) {
		    case DispRecCtrl_DRSoundRec_RecLevel: {
			XDispRecCtrl_DRSoundRec_RecLevelEvent
			    *reclevelEvent=
			    (XDispRecCtrl_DRSoundRec_RecLevelEvent*)drSoundRec_event;
			DispRecXtAPI_RecordingSession session = 
			    DispRecXtAPI_ConvertFrom_ExtRecordingSession(
				server, reclevelEvent->recordingSession, err 
			    );
			DispRecXtAPI_RecLevel reclevel;
			reclevel.fVeff = reclevelEvent->fVeff;
			reclevel.fOverdrivePpm = reclevelEvent->fOverdrivePpm;
			reclevel.fVeffSquare = reclevelEvent->fVeffSquare;
			reclevel.fMaxLevel = reclevelEvent->fMaxLevel;
			DispRecXtAPI_CallCallbacks(
			    DispRecXtAPI_RecordingSession, 
			    RecLevel,
			    session,
			    ( session, handle, reclevel, handle->fClientData )
			);	
			return True;
		    } break;
		}
	    } break;
	} /* end switch rpserv-event-type */
    }
    /* The event could not be dispatched -> return false */
    return False;
}

/*****************************************
Auxiliary functions
*****************************************/

/* memory management: Free arrays returned by API calls */
void DispRecXtAPI_FreeArray( void *array )
{
    DELETEARRAY( array );
}

/*****************************************
Init and Shutdown of API and connection to server 
*****************************************/

/**
* init this API and attach to the given app_context
*/
void DispRecXtAPI_Init( XtAppContext app_context, VtError *err )
{
    /* save app_context */
    DispRecXtAPI_app_context = app_context;
    
    {
	int i;
	for( i = 0; i < MAX_SERVERS; i++ ) {
	    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Server[i] = NULL;
	}
    }
}

/**
* shutdown this API
*/
void DispRecXtAPI_Shutdown( VtError *err )
{
    /* TODO */
}

/**
* connect to a DisplayRecorder/DemoRecorder X11-Proxy through the given Display.
* On error, DispRecXtAPI_Server_None is returned and the error-object set accordingly 
*/
DispRecXtAPI_Server DispRecXtAPI_Connect( Display* dpy, VtError *err )
{
    DispRecXtAPI_Server server = DispRecXtAPI_Server_None;
    
    if ( dpy == NULL ) {
	return DispRecXtAPI_Server_None;
    }
    
    server = NEW( struct DispRecXtAPI_Server_struct );
    
    server->fDisplay = dpy;
    
    /* init members */
    server->fSessions = DispRecXtAPI_RecordingSession_None;
    server->fGUIs = DispRecXtAPI_GUI_None;
    
    /* Query extension */
    {
	Bool haveDispRecCtrl;
	haveDispRecCtrl = 
	    XDispRecCtrl_QueryExtension(
		server->fDisplay, 
		&server->fDispRecCtrl_event_base, 
		&server->fDispRecCtrl_error_base
	    );	
	if (haveDispRecCtrl) {
	    /* Set the event mask */
	    XDispRecCtrl_SelectInput(
		server->fDisplay, 
		DispRecCtrl_NotifyMask | DispRecCtrl_DRSoundRecMask
	    );
	} else{
	    /* We cannot talk to rpserv --> set rpservDisplay to NULL */
	    DELETE( server );
	    return DispRecXtAPI_Server_None;
	}
    }
    
    /* determine fDispatchProcId */
    {
	int candidate = 0;
	for ( 
	    candidate = 0
	    ; 
	    ( DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Server[candidate] != NULL )
	    && 
	    ( candidate < MAX_SERVERS )
	    ; 
	    candidate++ 
	) {
	    /* do nothing */
	}
	if ( 
	    ( DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Server[candidate] == NULL )
	    && 
	    ( candidate < MAX_SERVERS )
	) {
	    /* found a valid candidate */
	    server->fDispatchProcId = candidate;
	    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Server[candidate] = server;
	} else {
	    /* didn't find a valid cndidate --> bail out with error */
	    /* TODO: set error object */
	    DELETE( server );
	    return DispRecXtAPI_Server_None;
	}
    }
    
    /* DispRecCtrl-Events: install hook into main-loop*/
    {
	int eventType;
	for ( 
	    eventType = server->fDispRecCtrl_event_base;
	    eventType < server->fDispRecCtrl_event_base+DispRecCtrl_NumberEvents;
	    eventType++
	) {
	    XtSetEventDispatcher(
		server->fDisplay, eventType, 
		DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Proc[server->fDispatchProcId]
	    );
	}
    }
    
    /* insert into list of servers */
    server->fNext = DispRecXtAPI_servers;
    if ( server->fNext ) {
	server->fNext->fPrev = &( server->fNext );
    }
    server->fPrev = &DispRecXtAPI_servers;
    DispRecXtAPI_servers = server;
    
    /* finished:  return the server object */
    return server;
}

/**
* Disconnect the server connection
*/
void DispRecXtAPI_Disconnect( 
    DispRecXtAPI_Server server, VtError *err )
{
    /* DispRecCtrl-Events: remove hook from  main-loop*/
    {
	int eventType;
	for ( 
	    eventType = server->fDispRecCtrl_event_base;
	    eventType < server->fDispRecCtrl_event_base+DispRecCtrl_NumberEvents;
	    eventType++
	) {
	    XtSetEventDispatcher(
		server->fDisplay, eventType, NULL);
	}
    }
    
    /* clean up */

    /* remove from DispatchProc */
    DispRecXtAPI_Dispatch_DispRecCtrlExtEvent_Server[ server->fDispatchProcId ] = NULL;
    server->fDispatchProcId = -1;

    /* remove from list of servers */
    *(server->fPrev) = server->fNext;
    if ( server->fNext ) {
	server->fNext->fPrev = server->fPrev;
    }
    server->fNext = NULL;
    server->fPrev = NULL;

    /* delete the server access data object */
    DELETE( server );
}


/*****************************************
Recording Sessions 
*****************************************/

/**
* Create a new recording session
*/
DispRecXtAPI_RecordingSession 
    DispRecXtAPI_CreateRecordingSession( 
	DispRecXtAPI_Server server, int flags, VtError *err 
    )
{
    DispRecXtAPI_RecordingSession session =
	NEW( struct DispRecXtAPI_RecordingSession_struct );
    session->fExtSession =
	XDispRecCtrl_CreateRecordingSession(
	    server->fDisplay,
	    flags
	);
    session->fServer = server;
    
    /* set data */
    
    /* TODO: change lower level api such that these properties are
	configured separately and not passed with starting of the recording
	anymore */
    session->fLength_recording_time = 0; /* 0 means: unlimited */
    session->fEndless = False;
    session->fDoSoundRecording = True;
    session->fMuteSoundRecording = False;
    session->fFinishOtherRecordingsBeforeStart = False;
    
    /* ownership --> True because we created it */
    session->fWeAreTheOwner = True;
    
    DispRecXtAPI_RecordingSession_InitDataAndInsertIntoStructure(
	server, session, err );
    
    /* return result */
    return session;
}


/**
* Delete a recording session
*/
void 
    DispRecXtAPI_DeleteRecordingSession( 
	DispRecXtAPI_RecordingSession session, VtError *err 
    )
{
    if ( session != DispRecXtAPI_RecordingSession_None ) {
	XDispRecCtrl_DeleteRecordingSession( 
	    session->fServer->fDisplay,  session->fExtSession 
	);
	
	DispRecXtAPI_Internal_DeleteRecordingSessionObject( session, err );
	
    }
}
    
/**
* Returns a list of all recording sessions of the given server.
* This includes recording sessions created by other clients.
*
* The list is returned as a pointer to an array of length *count.
* The size of the list is returned in "count".
*
* The list has to be deallocated by the caller using DispRecXtAPI_FreeArray();
*/
DispRecXtAPI_RecordingSession*
    DispRecXtAPI_GetAllRecordingSessions(
	DispRecXtAPI_Server server, int* count, VtError *err)
{
    /* TODO */
    *count = 0;
    return NULL;
}
	

/*****  Configure the recording session *****/

/* TODO */


/***** Start and Stop recording *****/

/**
* Sends a request to start recording to the given directory "dest_dir"
*
* Please note: 
*	This is an asynchronous request which means that
 *	an error-code OK does not confirm actual start of the recording
*	For this, please register and listen to the appropriate callbacks.
*/
void DispRecXtAPI_StartRecording(
    DispRecXtAPI_RecordingSession session,
    const char* dest_dir,
    enum DispRecCtrl_DirMode dir_mode,
    VtError *err )
{
    XDispRecCtrl_StartRecording(
	session->fServer->fDisplay,
	session->fExtSession,
	dest_dir,
	dir_mode,
	session->fLength_recording_time,
	session->fEndless,
	session->fDoSoundRecording,
	session->fMuteSoundRecording,
	session->fFinishOtherRecordingsBeforeStart
    );
}
    
void DispRecXtAPI_FinishRecording(
    DispRecXtAPI_RecordingSession session,
    VtError *err )
{
    XDispRecCtrl_FinishRecording(
	session->fServer->fDisplay,
	session->fExtSession
    );
}
    

/***** Query current state *****/

/**
* return True when the given session is currently recording 
*/
Bool DispRecXtAPI_Active(
    DispRecXtAPI_RecordingSession session,
    VtError *err )
{
    return 
	XDispRecCtrl_Active( 
	    session->fServer->fDisplay,
	    session->fExtSession
	);
}

/**
* return True if there is any recording session which is currently recording
*/
Bool DispRecXtAPI_AnyActive(
    DispRecXtAPI_Server server,
    VtError *err )
{
    /* TODO */
    return False;
}

/**
* Returns a list of all recording sessions of the given server which currently do recording
* This includes recording sessions created by other clients.
*
* The list is returned as a pointer to an array of length *count.
* The size of the list is returned in "count".
*
* As of now, not more than one session may be active at any time.
* But this may change in the future.
*
* The list has to be deallocated by the caller using DispRecXtAPI_FreeArray();
*/
DispRecXtAPI_RecordingSession*
    DispRecXtAPI_GetAllActiveRecordingSessions(
	DispRecXtAPI_Server server, int* count, VtError *err)
{
    /* TODO */
    *count = 0;
    return NULL;
}

/**
* If the recording session is controlled by a GUI, then return a handle to that GUI,
* otherwise return the DispRecXtAPI_GUI_None handle.
*/
DispRecXtAPI_GUI 
    DispRecXtAPI_GetGUIOfRecordingSession(
	DispRecXtAPI_RecordingSession session,
	VtError *err)
{
    /* TODO */
    return DispRecXtAPI_GUI_None;
}

/*****************************************
GUI
*****************************************/

/**
* Create a new GUI handle (GUI not started yet by this function)
*/
DispRecXtAPI_GUI
    DispRecXtAPI_CreateGUIHandle( 
	DispRecXtAPI_Server server, VtError *err 
    )
{
    DispRecXtAPI_GUI gui = NEW( struct DispRecXtAPI_GUI_struct );
    
    gui->fExtGui = 
	XDispRecCtrl_CreateGUIHandle(
	    server->fDisplay
	);
    if ( gui->fExtGui == DispRecCtrl_GUI_None ) {
	DELETE(gui);
	return DispRecXtAPI_GUI_None;
    }
    
    /* set data */
    DispRecXtAPI_Gui_InitDataAndInsertIntoStructure(
	server, gui, err );
    
    /* return result */
    return gui;
}
    
void DispRecXtAPI_DestroyGUIHandle(
    DispRecXtAPI_GUI gui, VtError *err )
{
    if ( gui != DispRecXtAPI_GUI_None ) {
	/* TODO: clean up */
	/* remove from list */
	*(gui->fPrev) = gui->fNext;
	if ( gui->fNext ) {
	    gui->fNext->fPrev = gui->fPrev;
	}
	gui->fNext = NULL;
	gui->fPrev = NULL;
	
	/* delete */
	DELETE( gui );
    }
}
    
/**
* Returns a list of all gui-handles of the given server.
* This includes gui-handles created by other clients.
*
* The list is returned as a pointer to an array of length *count.
* The size of the list is returned in "count".
*
* The list has to be deallocated by the caller using DispRecXtAPI_FreeArray();
*/
DispRecXtAPI_GUI*
    DispRecXtAPI_GetAllGUIs(
	DispRecXtAPI_Server server, int* count, VtError *err)
{
    /* TODO */
    *count = 0;
    return NULL;
}
	
/**** start and kill ****/

/**
* Request starting the GUI
* This request is handeled asynchronously, so be sure to add and listen to the appropriate callbacks
* for getting feedback about success or failure of the starting.
*
* The gui is started as the given user and group,
* This selecting of user and group only works if the DemoRecorder X11-proxy is run as root,
* otherwise user and group are ignored and the gui is run as the same user as the X11-proxy
*
* keepalive = True means that the GUI associated with that handle 
* is automatically restarted if it terminates.
*/
void DispRecXtAPI_StartGUI(
    DispRecXtAPI_GUI gui,
    uid_t user,
    gid_t group,
    Bool keepalive,    
    VtError *err)
{
    XDispRecCtrl_StartGUI(
	gui->fServer->fDisplay,
	gui->fExtGui,
	user, 
	group,
	keepalive
    );
}

/**
* Request starting the GUI if no GUI is already running.
* This request is handeled asynchronously, so be sure to add and listen to the appropriate callbacks
* for getting feedback about success or failure of the starting.
*
* if a GUI is already running, then if returns the DispRecXtAPI_GUI - handle of that GUI
* and sets the keepalive-property.
*
* The gui is started as the given user and group,
* This selecting of user and group only works if the DemoRecorder X11-proxy is run as root,
* otherwise user and group are ignored and the gui is run as the same user as the X11-proxy
*
* keepalive = True means that the GUI associated with that handle 
* is automatically restarted if it terminates.
*/
DispRecXtAPI_GUI DispRecXtAPI_StartGUIIfNotRunning(
    DispRecXtAPI_Server server,
    uid_t user,
    gid_t group,
    Bool keepalive,    
    VtError *err)
{
    DispRecXtAPI_GUI gui;
    DispRecCtrl_GUI ext_gui = 
	XDispRecCtrl_StartGUIIfNotRunning(
	    server->fDisplay,
	    user,
	    group,
	    keepalive);
    gui = 
	DispRecXtAPI_ConvertFrom_ExtGui(
	    server,
	    ext_gui,
	    err 
	);
    return gui;
}
  

/* set the keepalive property of a DispRecCtrl_GUI - handle.
    keepalive = True means that the GUI associated with that handle
    is automatically restarted if it terminates. */
void DispRecXtAPI_SetGUIKeepalive(
    DispRecXtAPI_Server server,
    DispRecXtAPI_GUI gui,
    Bool keepalive,
    VtError *err
) {
    XDispRecCtrl_SetGUIKeepalive(
	gui->fServer->fDisplay,
	gui->fExtGui,
	keepalive
    );
}

/**
* kill the given gui program. Implicitely disables keepalive.
*/
void DispRecXtAPI_KillGUI(
    DispRecXtAPI_GUI gui,
    VtError *err)
{
    XDispRecCtrl_KillGUI(
	gui->fServer->fDisplay,
	gui->fExtGui
    );
}


/****change visibility of gui ****/
/**
* make the gui visible and show it on the given screen  at the given coordinates
*/
void DispRecXtAPI_ShowGUI(
    DispRecXtAPI_GUI gui,
    int screen,
    int x, int y,
    VtError *err)
{
    XDispRecCtrl_ShowGUI(
	gui->fServer->fDisplay,
	gui->fExtGui,
	screen, x, y
    );
}

void DispRecXtAPI_HideGUI(
    DispRecXtAPI_GUI gui,
    VtError *err)
{
    XDispRecCtrl_HideGUI(
	gui->fServer->fDisplay,
	gui->fExtGui
    );
}
    
/**** query ****/

/** 
* returns True if there is a GUI program running under this GUI-Handle
*/
Bool DispRecXtAPI_GUIRunning( 
    DispRecXtAPI_GUI gui, VtError *err )
{
    return 
	XDispRecCtrl_GUIRunning(
	    gui->fServer->fDisplay,
	    gui->fExtGui
	);
}

/** 
* returns True if the GUI is visible
*/
Bool DispRecXtAPI_GUIVisible( 
    DispRecXtAPI_GUI gui_handle, VtError *err )
{
    return 
	XDispRecCtrl_GUIVisible(
	    gui_handle->fServer->fDisplay,
	    gui_handle->fExtGui
	);
}

/**
* if the GUI has a recording session, then return it, 
* otherwise return DispRecXtAPI_RecordingSession_None
*/
DispRecXtAPI_RecordingSession
     DispRecXtAPI_GetRecordingSessionOfGUI( 
	DispRecXtAPI_GUI gui, VtError *err
    )
{
    return 
	DispRecXtAPI_ConvertFrom_ExtRecordingSession(
	    gui->fServer,
	    XDispRecCtrl_GetRecordingSessionOfGUI(
		gui->fServer->fDisplay,
		gui->fExtGui
	    ),
	    err
	);
}
